require "util"
require "libs/array_pair"

placer = {}

require "config"


function string.starts_with(haystack, needle)
    return string.sub(haystack, 1, string.len(needle)) == needle
end


function string.ends_with(haystack, needle)
    return string.sub(haystack, -string.len(needle)) == needle
end


function math.round(x)
    return math.floor(x + .5)
end


function placer.init_globals()
    for index,_ in pairs(game.players) do
        placer.init_player(index)
    end
end


function placer.on_player_created(event)
    placer.init_player(event.player_index)
end


function placer.init_player(player_index)
    if not global.player_data then global.player_data = {} end

    local player_data = global.player_data[player_index]
    if not player_data then player_data = {} end
    if not player_data.overlays then player_data.overlays = {} end
    global.player_data[player_index] = player_data
end


local function make_ghost(surface, entity, params)
    params.name = 'entity-ghost'
    params.inner_name = params.inner_name or entity
    params.expires = false

    ---- Change to this to just make the things instead
    -- params.name = params.inner_name or entity

    surface.create_entity(params)
end


local function get_cardinal_direction(p, q)
    local radians = math.atan2(q.y - p.y, q.x - p.x)
    local quadrant = (math.round(radians / (math.pi / 2)) + 4) % 4
    local t = {[0]=defines.direction.east, [1]=defines.direction.south, [2]=defines.direction.west, [3]=defines.direction.north}
    return t[quadrant]
end


function placer.place_from_site(site)
    -- We'll build the belts running in the direction from the center of the
    -- site to the point that was clicked.
    local main_dir = get_cardinal_direction(site.center, site.start_pos)

    -- We use some indirection to handle all four possible directions with the
    -- same code. The variables are all named as if the direction is south: the
    -- belts run parallel to the "y" axis, positive or negative according to
    -- sgn.
    local sgn
    local X, Y
    local cross_dirs

    if main_dir == defines.direction.south then
        X = 'x'
        Y = 'y'
        sgn = 1
        cross_dirs = {[1]=defines.direction.east, [-1]=defines.direction.west}
    elseif main_dir == defines.direction.north then
        X = 'x'
        Y = 'y'
        sgn = -1
        cross_dirs = {[1]=defines.direction.west, [-1]=defines.direction.east}
    elseif main_dir == defines.direction.east then
        X = 'y'
        Y = 'x'
        sgn = 1
        cross_dirs = {[1]=defines.direction.south, [-1]=defines.direction.north}
    elseif main_dir == defines.direction.west then
        X = 'y'
        Y = 'x'
        sgn = -1
        cross_dirs = {[1]=defines.direction.north, [-1]=defines.direction.south}
    else
        return
    end

    local extents = {
        x={[-1]=math.floor(site.extents.left), [1]=math.floor(site.extents.right)},
        y={[-1]=math.floor(site.extents.top), [1]=math.floor(site.extents.bottom)},
    }

    -- Compute the points where we could place a miner such that it overlaps the
    -- site at all, by dilating the set of squares that have resources.
    local good_centers = {}
    for _, entity in pairs(site.entity_table) do
        local x = math.floor(entity[X])
        local y = math.floor(entity[Y])

        for x2 = x - 2, x + 2 do
            for y2 = y - 2, y + 2 do
                good_centers[string.format('%d-%d', x2, y2)] = 1
            end
        end
    end

    local surface = site.surface

    local y_extents = {}
    local min_miner_y, max_miner_y

    local belt_xs = {}

    -- Loop over output belts. We place the first belt at the "westmost"
    -- sensible location, and then tile from there. Perhaps we could try
    -- starting with the different offsets to see which leads to the fewest
    -- necessary miners.
    for x = extents[X][-sgn] + 4 * sgn, extents[X][sgn] + 4 * sgn, 7 * sgn do
        table.insert(belt_xs, x)
        -- Loop over possible positions along the belt, and place ghosts for
        -- those that would overlap the site at all.
        local y0, y1
        for y = extents[Y][-sgn] + 2 * sgn, extents[Y][sgn] + 2 * sgn, 3 * sgn do
            for _, d in pairs{1, -1} do
                local x2 = x - 2 * d * sgn
                if good_centers[string.format('%d-%d', x2, y)] then
                    make_ghost(surface, 'electric-mining-drill',
                               {
                                   direction = cross_dirs[d],
                                   position = {[X]=x2, [Y]=y},
                                   force = site.force,
                               }
                    )

                    if not y0 then
                        y0 = y
                    end
                    y1 = y
                end
            end
        end

        if not min_miner_y or y0 * sgn < min_miner_y * sgn then
            min_miner_y = y0
        end
        if not max_miner_y or y1 * sgn > max_miner_y * sgn then
            max_miner_y = y1
        end

        y_extents[x] = {y0, y1}
    end

    -- Now that we've done the miners, go back and place the belts. This is
    -- rather more complex, since we use a half-underground belt to fit electric
    -- poles, and have to handle some edge cases.

    -- Loop over the belt positions, same as before (but now with knowledge of
    -- the total range of miner positions).
    for _, x in ipairs(belt_xs) do
        -- Loop over possible y positions (locations along the belt). We might
        -- need one space before the first miner, to fit an electric pole, and
        -- up to two after, in case the last miner is on an input and we need to
        -- put a pole and an output after it.
        for y = y_extents[x][1] - sgn, max_miner_y + 2 * sgn, sgn do
            -- Our position along the cycle of pieces.
            local d = (sgn * (y - min_miner_y)) % 6

            local params
            if y * sgn > y_extents[x][2] * sgn then
                -- If we're past the last miner, we have special cases for the
                -- first two spaces when the last miner is on an input.
                -- Otherwise, we just lay down belts to match the longest row of
                -- miners along a belt.
                if y == y_extents[x][2] + sgn and d == 1 then
                    params = {
                        inner_name = 'small-electric-pole',
                    }
                elseif y == y_extents[x][2] + 2 * sgn and d == 2 then
                    params = {
                        inner_name = 'underground-belt',
                        type = 'output',
                    }
                else
                    params = {
                        inner_name = 'transport-belt',
                    }
                end
            else
                -- Within the range of miners, we mostly just cycle through the
                -- pieces normally. The exceptions are if the first miner would
                -- be an output belt, in which case we put use a regular belt,
                -- and making sure that we don't put an unnecessary belt when
                -- the first miner is on an input. (When the miner would be an
                -- output belt, we use the previous square for a pole.)
                if d == 0 then
                    params = {
                        inner_name = 'underground-belt',
                        type = 'input',
                    }
                elseif d == 2 then
                    params = {
                        inner_name = 'small-electric-pole',
                    }
                elseif d == 3 then
                    if y == y_extents[x][1] then
                        params = {
                            inner_name = 'transport-belt',
                        }
                    else
                        params = {
                            inner_name = 'underground-belt',
                            type = 'output',
                        }
                    end
                elseif d == 4 then
                    params = {
                        inner_name = 'transport-belt',
                    }
                elseif d == 5 then
                    if y * sgn >= y_extents[x][1] * sgn then
                        params = {
                            inner_name = 'transport-belt',
                        }
                    end
                end
            end

            if params then
                params.direction = main_dir
                params.position = {[X]=x, [Y]=y}
                params.force = site.force
                make_ghost(surface, nil, params)
            end
        end
    end

    -- Finally, place the extra belts to bring everything together.
    local n_belts = #belt_xs
    local center_x = (belt_xs[1] + belt_xs[n_belts]) / 2
    local left_x = center_x - (n_belts - 1) / 2 * sgn
    local min_across_y = max_miner_y + (2 + ((n_belts + 1) % 2) / 2) * sgn
    local max_across_y = min_across_y + (n_belts - 1) / 2 * sgn

    local p = game.players[1].print
    for i, x in ipairs(belt_xs) do
        local end_x = left_x + (i - 1) * sgn
        local across_y = min_across_y + math.abs(center_x - end_x) * sgn

        p(string.format('%d %d', end_x, across_y))

        for y = max_miner_y + 3 * sgn, across_y - sgn, sgn do
            make_ghost(surface, 'transport-belt',
                       {
                           direction = main_dir,
                           position = {[X]=x, [Y]=y},
                           force = site.force,
                       }
            )
        end
        for y = across_y, max_across_y, sgn do
            make_ghost(surface, 'transport-belt',
                       {
                           direction = main_dir,
                           position = {[X]=end_x, [Y]=y},
                           force = site.force,
                       }
            )
        end

        if end_x ~= x then
            local sgn2 = math.abs(end_x - x) / (end_x - x)
            for x2 = x, end_x - sgn2, sgn2 do
                make_ghost(surface, 'transport-belt',
                           {
                               direction = cross_dirs[sgn2 * sgn],
                               position = {[X]=x2, [Y]=across_y},
                               force = site.force,
                           }
                )
            end
        end
    end
end


local function position_to_string(entity)
    -- scale it up so (hopefully) any floating point component disappears,
    -- then force it to be an integer with %d.  not using util.positiontostr
    -- as it uses %g and keeps the floating point component.
    return string.format("%d,%d", entity.x * 100, entity.y * 100)
end


local function find_resource_at(surface, position)
    -- The position we get is centered in its tile (e.g., {8.5, 17.5}).
    -- Sometimes, the resource does not cover the center, so search the full tile.
    local top_left = {x = position.x - 0.5, y = position.y - 0.5}
    local bottom_right = {x = position.x + 0.5, y = position.y + 0.5}

    local stuff = surface.find_entities_filtered{area={top_left, bottom_right}, type='resource'}
    if #stuff < 1 then return nil end

    return stuff[1] -- there should never be another resource at the exact same coordinates
end


function placer.on_built_entity(event)
    if event.created_entity.name ~= 'mine-placer' then return end

    local player = game.players[event.player_index]
    local player_data = global.player_data[event.player_index]
    local pos = event.created_entity.position
    local surface = event.created_entity.surface

    -- Don't actually place the resource monitor entity
    if not player.cursor_stack.valid_for_read then
        player.cursor_stack.set_stack{name="mine-placer", count=1}
    elseif player.cursor_stack.name == "mine-placer" then
        player.cursor_stack.count = player.cursor_stack.count + 1
    end
    event.created_entity.destroy()

    local resource = find_resource_at(surface, pos)
    if not resource then
        -- if we have an expanding site, submit it. else, just drop the current site
        if player_data.current_site and player_data.current_site.is_site_expanding then
            placer.submit_site(event.player_index)
        else
            placer.clear_current_site(event.player_index)
        end
        return
    end

    local rescat = resource.prototype.resource_category
    if rescat and rescat == "sand" then
        player.print{"YARM-err-sand-is-bad", rescat}
        return
    end

    placer.add_resource(event.player_index, resource)
end


function placer.clear_current_site(player_index)
    local player = game.players[player_index]
    local player_data = global.player_data[player_index]

    player_data.current_site = nil

    while #player_data.overlays > 0 do
        table.remove(player_data.overlays).destroy()
    end
end


function placer.add_resource(player_index, entity)
    local player = game.players[player_index]
    local player_data = global.player_data[player_index]

    if player_data.current_site and player_data.current_site.ore_type ~= entity.name then
        if player_data.current_site.finalizing then
            placer.submit_site(player_index)
        else
            placer.clear_current_site(player_index)
        end
    end

    if not player_data.current_site then
        player_data.current_site = {
            added_at = game.tick,
            surface = entity.surface,
            force = player.force,
            ore_type = entity.name,
            ore_name = entity.prototype.localised_name,
            entity_table = {},
            entity_count = 0,
            initial_amount = 0,
            amount = 0,
            extents = {
                left = entity.position.x,
                right = entity.position.x,
                top = entity.position.y,
                bottom = entity.position.y,
            },
            next_to_scan = {},
            entities_to_be_overlaid = {},
            next_to_overlay = {},
            start_pos = {x=entity.position.x, y=entity.position.y}
        }
    end

    if player_data.current_site.is_site_expanding then
        player_data.current_site.has_expanded = true -- relevant for the console output
        if not player_data.current_site.original_amount then
            player_data.current_site.original_amount = player_data.current_site.amount
        end
    end

    placer.add_single_entity(player_index, entity)
    -- note: placer.scan_current_site() (via on_tick) will continue the operation from here
end


function placer.add_single_entity(player_index, entity)
    local player_data = global.player_data[player_index]
    local site = player_data.current_site
    local entity_pos = entity.position

    -- Don't re-add the same entity multiple times
    local key = position_to_string(entity_pos)
    if site.entity_table[key] then
        return
    end

    if site.finalizing then site.finalizing = false end

    -- Memorize this entity
    site.entity_table[key] = entity_pos
    site.entity_count = site.entity_count + 1
    table.insert(site.next_to_scan, entity)
    site.amount = site.amount + entity.amount

    -- Resize the site bounds if necessary
    if entity.position.x < site.extents.left then
        site.extents.left = entity.position.x
    elseif entity.position.x > site.extents.right then
        site.extents.right = entity.position.x
    end
    if entity.position.y < site.extents.top then
        site.extents.top = entity.position.y
    elseif entity.position.y > site.extents.bottom then
        site.extents.bottom = entity.position.y
    end

    -- Give visible feedback, too
    placer.put_marker_at(entity.surface, entity_pos, player_data)
end


function placer.put_marker_at(surface, pos, player_data)
    if math.floor(pos.x) % placer.overlay_step ~= 0 or
       math.floor(pos.y) % placer.overlay_step ~= 0 then
        return
    end

    local overlay = surface.create_entity{name="mp_overlay",
                                          force=game.forces.neutral,
                                          position=pos}
    overlay.minable = false
    overlay.destructible = false
    overlay.operable = false
    table.insert(player_data.overlays, overlay)
end


local function shift_position(position, direction)
    if direction == defines.direction.north then
        return {x = position.x, y = position.y - 1}
    elseif direction == defines.direction.northeast then
        return {x = position.x + 1, y = position.y - 1}
    elseif direction == defines.direction.east then
        return {x = position.x + 1, y = position.y}
    elseif direction == defines.direction.southeast then
        return {x = position.x + 1, y = position.y + 1}
    elseif direction == defines.direction.south then
        return {x = position.x, y = position.y + 1}
    elseif direction == defines.direction.southwest then
        return {x = position.x - 1, y = position.y + 1}
    elseif direction == defines.direction.west then
        return {x = position.x - 1, y = position.y}
    elseif direction == defines.direction.northwest then
        return {x = position.x - 1, y = position.y - 1}
    else
        return position
    end
end


function placer.scan_current_site(player_index)
    local site = global.player_data[player_index].current_site

    local to_scan = math.min(100, #site.next_to_scan)
    for i = 1, to_scan do
        local entity = table.remove(site.next_to_scan, 1)
        local entity_position = entity.position
        local surface = entity.surface

        -- Look in every direction around this entity...
        for _, dir in pairs(defines.direction) do
            -- ...and if there's a resource, add it
            local found = find_resource_at(surface, shift_position(entity_position, dir))
            if found and found.name == site.ore_type then
                placer.add_single_entity(player_index, found)
            end
        end
    end
end


local function find_center(area)
    local xpos = (area.left + area.right) / 2
    local ypos = (area.top + area.bottom) / 2

    return {x = math.floor(xpos),
            y = math.floor(ypos)}
end


function placer.finalize_site(player_index)
    local player = game.players[player_index]
    local player_data = global.player_data[player_index]

    local site = player_data.current_site
    site.finalizing = true
    site.finalizing_since = game.tick

    site.center = find_center(site.extents)

    --[[ don't rename a site we've expanded! (if the site name changes it'll create a new site
         instead of replacing the existing one) ]]
    if not site.is_site_expanding then
        site.name = 'name' --string.format("%s %d", get_octant_name(site.center), util.distance({x=0, y=0}, site.center))
    end

    --placer.count_deposits(site, site.added_at % placer.ticks_between_checks)
end


function placer.submit_site(player_index)
    local player = game.players[player_index]
    local player_data = global.player_data[player_index]
    -- local force_data = global.force_data[player.force.name]
    local site = player_data.current_site

    -- force_data.ore_sites[site.name] = site
    placer.clear_current_site(player_index)
    if (site.is_site_expanding) then
        if(site.has_expanded) then
            local amount_added = site.amount - site.original_amount
            local sign = amount_added < 0 and '' or '+' -- format_number will handle the negative sign for us (if needed)
        end
        --[[ NB: deliberately not outputting anything in the case where the player cancelled (or
             timed out) a site expansion without expanding anything (to avoid console spam) ]]
    end

    placer.place_from_site(site)

    -- clear site expanding state so we can re-expand the same site again (and get sensible numbers!)
    if(site.is_site_expanding) then
        site.is_site_expanding = nil
        site.has_expanded = nil
        site.original_amount = nil
    end
end


function placer.count_deposits(site, update_cycle)
    if site.iter_fn then
        placer.tick_deposit_count(site)
        return
    end

    local site_update_cycle = site.added_at % placer.ticks_between_checks
    if site_update_cycle ~= update_cycle then
        return
    end

    site.iter_fn, site.iter_state, site.iter_key = pairs(site.entity_table)
    site.update_amount = 0
end


function placer.start_recreate_overlay_existing_site(player_index)
    local site = global.player_data[player_index].current_site
    site.is_overlay_being_created = true

    -- forcible cleanup in case we got interrupted during a previous background overlay attempt
    site.entities_to_be_overlaid = {}
    site.entities_to_be_overlaid_count = 0
    site.next_to_overlay = {}
    site.next_to_overlay_count = 0

    for key,pos in pairs(site.entity_table) do
        local ent = site.surface.find_entity(site.ore_type, pos)
        if ent and ent.valid then
            site.entities_to_be_overlaid[key] = pos
            site.entities_to_be_overlaid_count = site.entities_to_be_overlaid_count + 1
        end
    end
end

function placer.process_overlay_for_existing_site(player_index)
    local player_data = global.player_data[player_index]
    local site = player_data.current_site

    if site.next_to_overlay_count == 0 then
        if site.entities_to_be_overlaid_count == 0 then
            placer.end_overlay_creation_for_existing_site(player_index)
            return
        else
            local ent_key, ent_pos = next(site.entities_to_be_overlaid)
            site.next_to_overlay[ent_key] = ent_pos
            site.next_to_overlay_count = site.next_to_overlay_count + 1
        end
    end

    local to_scan = math.min(100, site.next_to_overlay_count)
    for i = 1, to_scan do
        local ent_key, ent_pos = next(site.next_to_overlay)

        local entity = site.surface.find_entity(site.ore_type, ent_pos)
        local entity_position = entity.position
        local surface = entity.surface
        local key = position_to_string(entity_position)

        -- put marker down
        placer.put_marker_at(surface, entity_position, player_data)
        -- remove it from our to-do lists
        site.entities_to_be_overlaid[key] = nil
        site.entities_to_be_overlaid_count = site.entities_to_be_overlaid_count - 1
        site.next_to_overlay[key] = nil
        site.next_to_overlay_count = site.next_to_overlay_count - 1

        -- Look in every direction around this entity...
        for _, dir in pairs(defines.direction) do
            -- ...and if there's a resource that's not already overlaid, add it
            local found = find_resource_at(surface, shift_position(entity_position, dir))
            if found and found.name == site.ore_type then
                local offsetkey = position_to_string(found.position)
                if site.entities_to_be_overlaid[offsetkey] ~= nil and site.next_to_overlay[offsetkey] == nil then
                    site.next_to_overlay[offsetkey] = found.position
                    site.next_to_overlay_count = site.next_to_overlay_count + 1
                end
            end
        end
    end
end


function placer.end_overlay_creation_for_existing_site(player_index)
    local site = global.player_data[player_index].current_site
    site.is_overlay_being_created = false
    site.finalizing = true
    site.finalizing_since = game.tick
end


function placer.update_players(event)
    for index, player in pairs(game.players) do
        local player_data = global.player_data[index]

        if not player_data then
            placer.init_player(index)
        elseif not player.connected and player_data.current_site then
            placer.clear_current_site(index)
        end

        if player_data.current_site then
            local site = player_data.current_site

            if #site.next_to_scan > 0 then
                placer.scan_current_site(index)
            elseif not site.finalizing then
                placer.finalize_site(index)
            elseif site.finalizing_since + 10 == event.tick then
                placer.submit_site(index)
            end

            if site.is_overlay_being_created then
                placer.process_overlay_for_existing_site(index)
            end
        end
    end
end


function placer.on_tick(event)
    placer.update_players(event)
end
